html {
    position: relative;
    min-height: 100%;
}
body {
    /* Margin bottom by footer height */
    margin-bottom: 80px;
}
#footer {
    position: absolute;
    bottom: 0;
    width: 100%;
    /* Set the fixed height of the footer here */
    height: 80px;
    background-color: #f5f5f5;
}

a > code { color: #02d; }
a > code:hover { color: #44f; }
div.nav.bs-docs-sidenav > ul { list-style-type: none; }
div.nav.bs-docs-sidenav > ul > li > ul { list-style-type: none; }

h1, h2, h3, h4 {
    font-family: 'Roboto', 'sans-serif';
    margin-top: 40px;
}
p, div {
    font-family: 'Roboto Slab', 'sans-serif';
}

.container .text-muted {
    margin: 10px 0;
}

.jumboback > .navbar {
     background: rgba(0,0,0,0.4);
     margin-bottom: 0;
     text-shadow: none;
     border: 0;
     border-radius: 0;
}

.jumboback > .navbar > a, .jumboback > .navbar > ul > li > a {
    color: white;
}

.jumboback > .jumbotron {
    background: transparent;
}

.jumboback {
    background-image: url(background.jpg);
    margin-bottom: 20px;
    background-position: -20% 25%;
    background-size: cover;
    background-repeat: no-repeat;
    color: white;
    text-shadow: black 0em 0em 0.8em;
}
